#include "circuit.h"

int Circuit::check_circuit() {

    int sum_value_unsatisfied_cost = 0;

    int sum_fault_propagated_weight = 0;
    int sum_fault_propagated_unsatisfied_cost = 0;

    int sum_fault_detected_weight = 0;
    int sum_fault_detected_unsatisfied_cost = 0;

    for(Gate* g : gates) {

        assert(g->value_satisfied == ( g->cal_value() == g->value ));

        assert(g->fault_detected_satisfied[0] == ( g->cal_fault_detected(0) == g->fault_detected[0] ));
        assert(g->fault_detected_satisfied[1] == ( g->cal_fault_detected(1) == g->fault_detected[1] ));

        assert(g->fault_propagated_satisfied[0] == ( g->cal_propagate_len(0) == g->fault_propagated_len[0]));
        assert(g->fault_propagated_satisfied[1] == ( g->cal_propagate_len(1) == g->fault_propagated_len[1]));


        if(g->value_satisfied) {
            assert(value_satisfied_vars.count(g));
            assert(!value_unsatisfied_vars.count(g));
        } else {
            assert(value_unsatisfied_vars.count(g));
            assert(!value_satisfied_vars.count(g));
        }

        if(g->fault_detected_satisfied[0] && g->fault_detected_satisfied[1]) {
            assert(detected_satisfied_vars.count(g));
            assert(!detected_unsatisfied_vars.count(g));
        } else {
            assert(detected_unsatisfied_vars.count(g));
            assert(!detected_satisfied_vars.count(g));
        }

        if(g->fault_propagated_satisfied[0] && g->fault_propagated_satisfied[1]) {
            assert(propagated_satisfied_vars.count(g));
            assert(!propagated_unsatisfied_vars.count(g));
        } else {
            assert(propagated_unsatisfied_vars.count(g));
            assert(!propagated_satisfied_vars.count(g));
        }


        if(!g->value_satisfied) sum_value_unsatisfied_cost += g->value_unsatisfied_cost;

        if(g->fault_detected[0]) sum_fault_detected_weight += g->fault_detected_weight[0];
        if(g->fault_detected[1]) sum_fault_detected_weight += g->fault_detected_weight[1];

        if(!g->fault_detected_satisfied[0]) sum_fault_detected_unsatisfied_cost += g->fault_detected_unsatisfied_cost[0];
        if(!g->fault_detected_satisfied[1]) sum_fault_detected_unsatisfied_cost += g->fault_detected_unsatisfied_cost[1];

        sum_fault_propagated_weight += g->fault_propagated_len[0] * g->fault_propagated_weight[0];
        sum_fault_propagated_weight += g->fault_propagated_len[1] * g->fault_propagated_weight[1];

        if(!g->fault_propagated_satisfied[0]) sum_fault_propagated_unsatisfied_cost += g->fault_propagated_unsatisfied_cost[0];
        if(!g->fault_propagated_satisfied[1]) sum_fault_propagated_unsatisfied_cost += g->fault_propagated_unsatisfied_cost[1];
    }

    int circuit_score = sum_fault_propagated_weight + sum_fault_detected_weight
        - sum_fault_propagated_unsatisfied_cost - sum_fault_detected_unsatisfied_cost - sum_value_unsatisfied_cost;

    printf("svuc: %d, sfpw: %d, sfpuc: %d, sfdw: %d, sfduc: %d\n", 
        sum_value_unsatisfied_cost, sum_fault_propagated_weight, sum_fault_propagated_unsatisfied_cost,
        sum_fault_detected_weight, sum_fault_detected_unsatisfied_cost);

    return circuit_score;
}