#include "circuit.h"

#include <queue>
#include <unordered_set>
#include <unordered_map>
#include <algorithm>
#include "assert.h"
#include <chrono>


bool Circuit::local_search(std::unordered_set<Fault*> &faults) {

    // 随机生成初始电路
    ls_init_circuit(faults);

    // print_gates();

    printf("generate random circuit done!\n");
    int T = 10000000;

    int iter = 0;
    while(T--) {

        iter++;
        if(iter % 1000 == 0) {
            printf("iter: %d\n", iter);
            check_circuit();
        }

        if(value_unsatisfied_vars.empty()) {
            printf("Get Solution!\n");
            break;
        }

        Gate* gate = ls_pick();
        ls_flip(gate);

        // printf("svuc: %d, sfpw: %d, sfpuc: %d, sfdw: %d, sfduc: %d  <<<\n", 
        // flip_gate->score_value_unsatisfied_cost, 
        // flip_gate->score_fault_propagated_weight[0] + flip_gate->score_fault_propagated_weight[1], 
        // flip_gate->score_fault_propagated_unsatisfied_cost[0] + flip_gate->score_fault_propagated_unsatisfied_cost[1],
        // flip_gate->score_fault_detected_weight[0] + flip_gate->score_fault_detected_weight[1],
        // flip_gate->score_fault_detected_unsatisfied_cost[0] + flip_gate->score_fault_detected_unsatisfied_cost[1]);

        

        // int score_2 = check_circuit();
        // printf("checking circuit: %d %d\n", score_2 - score_1, dert);

        // assert((score_2 - score_1) ==  dert);
        // printf("i: %s\n", flip_gate->name.c_str());
    }

    // execute simulation
    for(Gate* g : rtopo_gates) {
        g->update_gate_statistics();
    }

    static int original_faults = -1;
    if(original_faults == - 1) {
        original_faults = faults.size();
    }
    static int pattern = 0;

    std::unordered_set<Fault*> tmp = faults;

    for(Fault* f : tmp) {
        if(f->gate->fault_detected[f->type]) {
            faults.erase(f);
        }
    }

    if(tmp.size() == faults.size()) pattern--;

    printf("coverage: %.3f%%\tpattern: %d\tbefore: %d\tnow: %d\n", (double)(original_faults - faults.size()) / (original_faults) * 100, ++pattern, tmp.size(), faults.size());

    exit(0);

    if(tmp.size() == faults.size()) return false;

    return true;
}

Gate* Circuit::ls_pick() {
    static int focus = 1;

    double SP = 0.1;

    if(rand() % 1000 <= 500 * SP) {
        std::vector<Gate*> tmp;
        for(Gate* g : value_unsatisfied_vars) {
            tmp.push_back(g);
        }
        return tmp[rand()%tmp.size()];
    }

    if(good_vars.size()) {
        return good_vars[rand()%good_vars.size()];
    }

    // printf("update weight\n");
    ls_update_weight();

    std::vector<Gate*> tmp;
    for(Gate* g : value_unsatisfied_vars) {
        tmp.push_back(g);
    }

    return tmp[rand()%tmp.size()];
}

void Circuit::ls_update_weight() {

    std::unordered_set<Gate*> gates_need_recal_score;

    for(Gate* g : value_unsatisfied_vars) {
        g->value_unsatisfied_cost += 1;
        for(Gate *in : g->fanins) {
            gates_need_recal_score.insert(in);
        }
        for(Gate *out : g->fanouts) {
            gates_need_recal_score.insert(out);
        }
    }

    for(Gate* g : detected_unsatisfied_vars) {
        g->fault_detected_unsatisfied_cost[0] += g->level;
        g->fault_detected_unsatisfied_cost[1] += g->level;
        for(Gate* r : g->score_reigon) {
            gates_need_recal_score.insert(r);
        }
    }

    for(Gate* g : propagated_unsatisfied_vars) {
        g->fault_propagated_unsatisfied_cost[0] += g->level;
        g->fault_propagated_unsatisfied_cost[1] += g->level;
        for(Gate* r : g->score_reigon) {
            gates_need_recal_score.insert(r);
        }
    }

    for(Gate *g : gates_need_recal_score) {
        g->cal_score();
    }

    // check_circuit();

}

void Circuit::ls_flip(Gate* gate) {

    // printf("flip: %s\n", gate->name.c_str());

    gate->value ^= 1;
    gate->update_gate_property();
    gate->update_gate_statistics();

    for(Gate* g : gate->fanouts) {
        g->update_gate_statistics();
    }
    for(Gate* g : gate->reigon) {
        g->update_gate_statistics();
    }

    // for(Gate *g : gates) {
    //     g->score = g->cal_score();
    // }

    gate->cal_score();

    for(Gate* g : gate->score_reigon) {
        g->cal_score();
    }

    // printf("t: %d cal_cnt: %d   r_set: %d\n", t.size(), cal_cnt, gate->score_reigon.size());

    // for(Gate* g : gates) {
    //     if(g->score != g->cal_score(true) ){
    //         printf("bug3: %s\n", g->name.c_str());
    //         // print_gates();
    //         exit(0);
    //     }
    // }

    // exit(0);

    // printf("bug %s: cal_len:%d len:%d\n", gate->name.c_str(), gate->cal_propagate_len(1), gate->fault_propagated_len[1]);
}

void Circuit::ls_init_circuit(std::unordered_set<Fault*> &faults) {

    good_vars.clear();
    value_unsatisfied_vars.clear();
    value_satisfied_vars.clear();
    propagated_unsatisfied_vars.clear();
    propagated_satisfied_vars.clear();
    detected_unsatisfied_vars.clear();
    detected_satisfied_vars.clear();

    // init value, weight and cost
    for(Gate* g : gates) {
        g->value = rand() % 2;
        g->value_unsatisfied_cost = 1;
        g->fault_propagated_weight[0] = g->fault_propagated_weight[1] = 0;
        g->fault_propagated_unsatisfied_cost[0] = g->fault_propagated_unsatisfied_cost[1] = 1;
        g->fault_detected_weight[0] = g->fault_detected_weight[1] = 0;
        g->fault_detected_unsatisfied_cost[0] = g->fault_detected_unsatisfied_cost[1] = 1;
        g->good_var_index = -1;
        g->unsatisfied_var_index = -1;
    }

    for(Fault* f : faults) {
        f->gate->fault_propagated_weight[f->type] = 1;
        f->gate->fault_detected_weight[f->type] = 1;
    }

    for(Gate* g : rtopo_gates) {
        g->update_gate_property();
    }

    for(Gate* g : rtopo_gates) {
        g->update_gate_statistics();
    }

    // cal score
    for(Gate* g : gates) {
        g->cal_score();
        printf("%d ", g->score);
    }

}