#include "circuit.h"


void Gate::cal_score(int debug) {

    value ^= 1;

    if(debug) {
        // printf("lala: %s\n", this->name.c_str());
        assert(score_value_unsatisfied_cost == cal_value_unsatisfied_cost());
        assert(score_fault_propagated_weight[0] == cal_score_fault_propagated_weight(0));
        assert(score_fault_propagated_weight[1] == cal_score_fault_propagated_weight(1));
        assert(score_fault_propagated_unsatisfied_cost[0] == cal_score_fault_propagated_unsatisfied_cost(0));
        assert(score_fault_propagated_unsatisfied_cost[1] == cal_score_fault_propagated_unsatisfied_cost(1));
        assert(score_fault_detected_weight[0] == cal_score_fault_detected_weight(0));
        assert(score_fault_detected_weight[1] == cal_score_fault_detected_weight(1));
        assert(score_fault_detected_unsatisfied_cost[0] == cal_score_fault_detected_unsatisfied_cost(0));
        assert(score_fault_detected_unsatisfied_cost[1] == cal_score_fault_detected_unsatisfied_cost(1));
    }

    score_value_unsatisfied_cost = cal_value_unsatisfied_cost();
    score_fault_propagated_weight[0] = cal_score_fault_propagated_weight(0);
    score_fault_propagated_weight[1] = cal_score_fault_propagated_weight(1);

    score_fault_propagated_unsatisfied_cost[0] = cal_score_fault_propagated_unsatisfied_cost(0);
    score_fault_propagated_unsatisfied_cost[1] = cal_score_fault_propagated_unsatisfied_cost(1);

    score_fault_detected_weight[0] = cal_score_fault_detected_weight(0);
    score_fault_detected_weight[1] = cal_score_fault_detected_weight(1);

    score_fault_detected_unsatisfied_cost[0] = cal_score_fault_detected_unsatisfied_cost(0);
    score_fault_detected_unsatisfied_cost[1] = cal_score_fault_detected_unsatisfied_cost(1);

    value ^= 1;

    score = - score_value_unsatisfied_cost 
    + score_fault_propagated_weight[0] + score_fault_propagated_weight[1]
    - score_fault_propagated_unsatisfied_cost[0] - score_fault_propagated_unsatisfied_cost[1]
    + score_fault_detected_weight[0] + score_fault_detected_weight[1]
    - score_fault_detected_unsatisfied_cost[0] - score_fault_detected_unsatisfied_cost[1];

    if(score > 0 && good_var_index == -1) {
        good_var_index = C->good_vars.size();
        C->good_vars.push_back(this);
    }

    if(score <= 0 && good_var_index != -1) {
        C->good_vars[C->good_vars.size()-1]->good_var_index = good_var_index;
        std::swap(C->good_vars[good_var_index], C->good_vars[C->good_vars.size()-1]);
        C->good_vars.pop_back();
        good_var_index = -1;
    }

    for(Gate* g : C->good_vars) {
        assert(g->score > 0);
    }

    return;
}

int Gate::cal_value_unsatisfied_cost() {
    int res = 0;
    for(Gate* g : fanouts) {
        if(g->value_satisfied && g->cal_value() != g->value) {
            res += g->value_unsatisfied_cost;
        }
        if(!g->value_satisfied && g->cal_value() == g->value) {
            res -= g->value_unsatisfied_cost;
        }
    }
    if(this->value_satisfied && this->cal_value() != this->value) {
        res += this->value_unsatisfied_cost;
    }
    if(!this->value_satisfied && this->cal_value() == this->value) {
        res -= this->value_unsatisfied_cost;
    }
    return res;
}

int Gate::cal_score_fault_propagated_weight(int sa) {
    int res = 0;
    res += (this->cal_propagate_len(sa) - this->fault_propagated_len[sa] ) * this->fault_propagated_weight[sa];
    return res;
}

int Gate::cal_score_fault_propagated_unsatisfied_cost(int sa) {
    int res = 0;
    int origin_fault_propagated_len[2];
    origin_fault_propagated_len[0] = this->fault_propagated_len[0];
    origin_fault_propagated_len[1] = this->fault_propagated_len[1];
    this->fault_propagated_len[0] = this->cal_propagate_len(0);
    this->fault_propagated_len[1] = this->cal_propagate_len(1);

    if(!this->fault_propagated_satisfied[sa]) {
        res -= this->fault_propagated_unsatisfied_cost[sa];
    }

    for(Gate* g : this->reigon) {
        if(g->fault_propagated_satisfied[sa] && g->cal_propagate_len(sa) != g->fault_propagated_len[sa]) {
            res += g->fault_propagated_unsatisfied_cost[sa];
        }
        if(!g->fault_propagated_satisfied[sa] && g->cal_propagate_len(sa) == g->fault_propagated_len[sa]) {
            res -= g->fault_propagated_unsatisfied_cost[sa];
        }
    }

    this->fault_propagated_len[0] = origin_fault_propagated_len[0];
    this->fault_propagated_len[1] = origin_fault_propagated_len[1];
    return res;
}

int Gate::cal_score_fault_detected_weight(int sa) {
    int res = 0;
    res += ( this->cal_fault_detected(sa) - this->fault_detected[sa] ) * this->fault_detected_weight[sa];
    return res;
}

int Gate::cal_score_fault_detected_unsatisfied_cost(int sa) {
    int res = 0;

    int origin_fault_detected[2];
    origin_fault_detected[0] = this->fault_detected[0];
    origin_fault_detected[1] = this->fault_detected[1];
    this->fault_detected[0] = this->cal_fault_detected(0);
    this->fault_detected[1] = this->cal_fault_detected(1);

    if(!fault_detected_satisfied[sa]) {
        res -= this->fault_detected_unsatisfied_cost[sa];
    }

    for(Gate* g : this->reigon) {
        if(g->fault_detected_satisfied[sa] && g->cal_fault_detected(sa) != g->fault_detected[sa]) {
            res += g->fault_detected_unsatisfied_cost[sa];
        }
        if(!g->fault_detected_satisfied[sa] && g->cal_fault_detected(sa) == g->fault_detected[sa]) {
            res -= g->fault_detected_unsatisfied_cost[sa];
        }
    }

    this->fault_detected[0] = origin_fault_detected[0];
    this->fault_detected[1] = origin_fault_detected[1];

    return res;
}